

#    MERGE run control 

global env

global Options

variable MaxTries 20

variable SYS_DataAcquisitionCode  0
variable SYS_PrevDataAcquisitionCode  0
variable SYS_DataAcquisitionArgs 0

variable RunControlDelay1 100
variable RunControlDelay2 50
variable RunControlDelay3 500

variable NLinks 64    ;#   default

#    list of registers required by the UI

variable registerlist

set registerlist {}

lappend registerlist [list NetVar.MERGE.Action                "-n MERGE_Action -t int"]
lappend registerlist [list NetVar.MERGE.Signal                "-n MERGE_Signal -t int"]
lappend registerlist [list NetVar.MERGE.Parameter             "-n MERGE_Parameter -t string"]
lappend registerlist [list NetVar.MERGE.ID                    "-n MERGE_ID -t string"]
lappend registerlist [list NetVar.MERGE.Response              "-n MERGE_Response -t block -a yes -i 0 -l 444 -f x"]

lappend registerlist [list NetVar.MERGE.PID                   "-n MERGE_PID -t int"]
lappend registerlist [list NetVar.MERGE.State                 "-n MERGE_State -t int"]
lappend registerlist [list NetVar.MERGE.Paused                "-n MERGE_Paused -t int"]
lappend registerlist [list NetVar.MERGE.Trace                 "-n MERGE_Trace -t int"]
lappend registerlist [list NetVar.MERGE.RunOptions            "-n MERGE_RunOptions -t int"]
lappend registerlist [list NetVar.MERGE.HardwareVersion       "-n MERGE_HardwareVersion -t int"]
lappend registerlist [list NetVar.MERGE.Verbose               "-n MERGE_Verbose -t int"]
lappend registerlist [list NetVar.MERGE.SyncStep              "-n MERGE_SyncStep -t int"]

lappend registerlist [list NetVar.MERGE.TestMode              "-n MERGE_TestMode -t int"]
lappend registerlist [list NetVar.MERGE.PollTicks             "-n MERGE_PollTicks -t int"]
lappend registerlist [list NetVar.MERGE.InitialTimeStamp      "-n MERGE_InitialTimeStamp -t int"]
lappend registerlist [list NetVar.MERGE.MsgLevel              "-n MERGE_MsgLevel -t int -f x"]

# for the LINK task

lappend registerlist [list NetVar.MERGE.LinksAvailable     "-n MERGE_LinksAvailable -t int -f d"]
lappend registerlist [list NetVar.MERGE.Links.Alive        "-n MERGE_LinksAlive -t int -f d"]
lappend registerlist [list NetVar.MERGE.Links.InUse        "-n MERGE_LinksInUse -t string"]

lappend registerlist [list NetVar.MERGE.Link.Alive         "-n MERGE_LinkAlive   -t block -a yes -i 0 -l 512 -f d"]
lappend registerlist [list NetVar.MERGE.Link.InUse         "-n MERGE_LinkInUse   -t block -a yes -i 0 -l 512 -f d"]
lappend registerlist [list NetVar.MERGE.Link.Ctrl          "-n MERGE_LinkCtrl    -t block -a yes -i 0 -l 512 -f d"]
lappend registerlist [list NetVar.MERGE.Link.State         "-n MERGE_LinkState   -t block -a yes -i 0 -l 512 -f d"]
lappend registerlist [list NetVar.MERGE.Link.Status        "-n MERGE_LinkStatus  -t block -a yes -i 0 -l 512 -f d"]
lappend registerlist [list NetVar.MERGE.Link.TcpPort       "-n MERGE_LinkTcpPort -t block -a yes -i 0 -l 512 -f d"]

#    Stats for MERGE

lappend registerlist [list NetVar.MERGE.StatsTime         "-n MERGE_StatsTime -t int"]
lappend registerlist [list NetVar.MERGE.ZeroStats         "-n MERGE_ZeroStats -t int"]

for {set i 0} {$i <= $NLinks} {incr i} {
    lappend registerlist [list NetVar.MERGE.Stats$i  "-n MERGE_Stats$i -t block -a yes -i 0 -l 512 -f d"]
    lappend registerlist [list NetVar.MERGE.Rates$i  "-n MERGE_Rates$i -t block -a yes -i 0 -l 512 -f d"]
}


proc make-UIregisters {} {
   variable registerlist

    set rc [catch {ReadNetVar NetVar.MERGE.ID} m]
    if {$rc == 0} return
    if {$rc == 1 || $m == 0x10004} {
        foreach item $registerlist {
            set rc [catch {CreateNetVar [first $item] [second $item]} m]
        }
        set rc1 [catch {CreateSigTask SigTask.MERGE.Usr1 "-t MERGE -s USR1"} m1]
        set rc2 [catch {CreateSigTask SigTask.MERGE.Usr2 "-t MERGE -s USR2"} m2]
        puts "Created UI registers"
    } else {
        puts "make-UIregisters returns rc=$rc and $m"
    }
}


proc MergeCommand {args} {
    variable MaxTries
    variable RunControlDelay2
    variable RunControlDelay3

#    first      action
#    second     message
#    third      parameter
#    fourth     signal number
#    fifth      task name

#   default args eg   DataAcqCommand 1 "Halt" 0 2 MERGE

#    step 1  setup action and send message to task 

    if {[llength $args] > 4} then {set TASK [fifth $args]} else {set TASK MERGE}

        set pid [ReadNetVar NetVar.$TASK.PID]
        if {[llength $args] > 2} {WriteNetVar NetVar.$TASK.Parameter [third $args]}
        WriteNetVar NetVar.$TASK.Action [first $args]
        if {[llength $args] > 3} {
            WriteSigTaskAttr SigTask.$TASK.Usr[fourth $args] "-t $pid"
            WriteSigTask SigTask.$TASK.Usr[fourth $args] 0
        } else {
            WriteSigTaskAttr SigTask.$TASK.Usr2 "-t $pid"
            WriteSigTask SigTask.$TASK.Usr2 0
        }

        after $RunControlDelay2

#    step 2  wait for action to be completed and response received 

     if {[llength $args] > 1} {set msg "[second $args]"} else {set msg ""}

    set tries 1

    while {$tries < $MaxTries} {

          if {[llength $args] == 0} then {
              puts "Waiting for action to complete ($tries)"
          } else {
             if {[llength $args] > 1} then {
                puts "[second $args] ($tries)"
             } else {
                puts "[first $args] ($tries)"
             }
          }

        set z [catch {ReadNetVar NetVar.$TASK.Action} m ]

        if {$z != 0} then {
            puts "Unable to confirm state of action\n$m"
            return 1
        }

	  if {$m == 0} then {puts "Action has completed: MergeCommand $args"; return 0}

        incr tries
        after $RunControlDelay3
    }

    puts "Action has not completed\nGiving up waiting"
    return 1
}

    proc MergeState {} {

       set state [expr [ReadNetVar NetVar.MERGE.State] - 1]
       if {$state < 0 || $state > 9} {set state 8}

       return $state
    }

    proc Set_Options {} {
        global Options

        LoadOptions

        if {![info exists Options(MERGE.LinksAvailable)]} {set Options(MERGE.LinksAvailable) 4}
        if {![info exists Options(MERGE.LinksInUse)]} {set Options(MERGE.LinksInUse) "1%1%1%1%"}
        if {![info exists Options(MERGE.RunOptions)]} {set Options(MERGE.RunOptions) 1}

        SaveOptions
    }

    proc load_MergeLinks {} {
       global Options
       variable NLinks
       variable NRows
       variable NCols
       variable LinksAvailable
       variable LinksInUse
       variable LinkInUse

       Set_Options; make-UIregisters

       set LinksAvailable $Options(MERGE.LinksAvailable)
       WriteNetVar NetVar.MERGE.LinksAvailable $Options(MERGE.LinksAvailable)

       if {$LinksAvailable <= 8} {
         set NRows 1
         set NCols $LinksAvailable
       } else {
         set NRows [expr ($LinksAvailable+7) / 8]
         set NCols 8
       }
       set NLinks [expr $NRows * $NCols]

       set LinksInUse [split $Options(MERGE.LinksInUse) %]
       WriteNetVar NetVar.MERGE.Links.InUse $Options(MERGE.LinksInUse)
       
       for {set link 0} {$link < $NLinks} {incr link} {
          if {$link < [expr [llength $LinksInUse] - 1]} {
             set LinkInUse($link) [lindex $LinksInUse $link]
          } else {
             set LinkInUse($link) 0
          }
       }
    }

    proc store_MergeLinks {} {
       global Options
       variable LinksAvailable
       variable LinksInUse
       variable NLinks
       variable LinkInUse

       Set_Options; make-UIregisters

       set LinksInUse ""

       for {set link 0} {$link < $NLinks} {incr link} {
          append LinksInUse $LinkInUse($link)
          append LinksInUse "%"
       }

       set Options(MERGE.LinksAvailable) $LinksAvailable
       set Options(MERGE.LinksInUse) $LinksInUse

       WriteNetVar NetVar.MERGE.Links.InUse $Options(MERGE.LinksInUse)

       SaveOptions
    }

    proc set_MergeLinks {} {
       variable LinksInUse

       WriteNetVar NetVar.MERGE.Links.InUse $LinksInUse
    }


proc doMerge {vl} {
    variable Options

puts "MERGE command  $vl"

#    0=>11    control MERGE task
#    20=>22   control LINK task
#    30=>35   control overall NewMerger

    Set_Options; make-UIregisters

    switch $vl {
        0  {doResetMerge}
        1  {doStopMerge}
        2  {doGoMerge}
        3  {doFlushMerge}
        4  {doSetupMerge}
        5  {doPauseMerge}
        6  {doResumeMerge}
        10 {doZeroTimeStamp}
        11 {doToggleXfer}
        20 {doResetLinks}
        21 {doStopLinks}
        22 {doGoLinks}
        30 {doSetUpMerger}
        31 {doStopMerger}
        32 {doGoMerger}
        33 {doResetMerger}
        34 {doPauseMerger}
        35 {doResumeMerger}
        default {}
    }

}

#    MERGE Commands
#define DO_RESET 0
#define DO_STOP 1
#define DO_GO 2
#define DO_SETUP 3
#define DO_RESET 4
#define DO_START 5
#define DO_PAUSE 6
#define DO_RESUME 7
#define doZeroTimeStamp 10
#define doToggleXfer 11

proc doResetMerge {} {puts "Reset Merge"; MergeCommand 4 "Reset"}

proc doSetupMerge {} {puts "Setup Merge"; MergeCommand 1 "Setup"}

proc doStopMerge {} {puts "Stop Merge"; MergeCommand 2 "Halt"}

proc doGoMerge {} {puts "Go Merge"; MergeCommand 3 "Go"}

proc doFlushMerge {} {puts "Flush Merge"; InsertLog "This operation is not yet in use / implemented"}

proc doPauseMerge {} {puts "Pause Merge"; InsertLog "This operation is not yet in use / implemented"}

proc doResumeMerge {} {puts "Resume Merge"; InsertLog "This operation is not yet in use / implemented"}

   proc doZeroTimeStamp {} {MergeCommand 10 "Zero Merge Time Stamp"}

   proc doToggleXfer {} {

      set RunOptions [ReadNetVar NetVar.MERGE.RunOptions]

      set Xbit [expr $RunOptions & 2]
      if {$Xbit == 0} then {set Xbit 2} else {set Xbit 0}
      set RunOptions [expr ($RunOptions & 0xfd) | $Xbit]

      WriteNetVar NetVar.MERGE.RunOptions $RunOptions
   }

#    commands for the MERGER
#define DO_STOP 31
#define DO_GO 32
#define DO_SETUP 30
#define DO_RESET 33
#define DO_PAUSE 34
#define DO_RESUME 35

proc doSetUpMerger {} {
    variable Options

puts "setup MERGER"

    Set_Options; make-UIregisters

    doStopMerge

#    Setup_MergeLinks

    doSetupMerge
}

proc doResetMerger {} {

puts "reset MERGER"

     Set_Options; make-UIregisters

     doResetMerge
     
#     Reset_MergeLinks
}

proc doStopMerger {} {

puts "stop MERGER"

     Set_Options; make-UIregisters

     doPauseMerger

     doStopLinks

     doStopMerge

     if {[expr [ReadNetVar NetVar.MERGE.RunOptions] & 16] == 16}  {
         MergeCommand 11 "disconnect xfer stream"
     }

    WriteNetVar NetVar.MERGE.Paused 1   ;#  FALSE
    WriteNetVar NetVar.MERGE.State 1    ;#  MERGE_STOPPED
}


proc doGoMerger {} {

puts "go MERGER"

     Set_Options; make-UIregisters

     if {[expr [ReadNetVar NetVar.MERGE.RunOptions] & 16] == 16}  {
         MergeCommand 12 "connect xfer stream"
     }

     doGoMerge

     doGoLinks 

     doResumeMerger

     WriteNetVar NetVar.MERGE.State 2     ;# MERGE_GOING
     WriteNetVar NetVar.MERGE.Paused 0    ;# TRUE

}

proc doPauseMerger {} {

puts "Pause MERGER"

  InsertLog "Pause Merger: This operation is not yet in use / implemented"

}

proc doResumeMerger {} {

puts "Resume MERGER"

  InsertLog "Resume MERGER: This operation is not yet in use / implemented"

}


#    commands for the Links
#define ResetLinks 20
#define StopLinks 21
#define GoLinks 22

proc doResetLinks {} {

   puts "Reset Links"

    InsertLog "Reset Links: This operation is not yet in use / implemented"
}

proc doStopLinks {} {
       variable NLinks
       variable LinksAvailable
       variable LinkInUse

       for {set link 0} {$link < $NLinks} {incr link} {
           if {$link < $LinksAvailable} then {
              if {$LinkInUse($link) == 1} then {
                 MergeCommand 21 "link $link STOP" $link 2
              }
           }
        }
}

proc doGoLinks {} {
       variable NLinks
       variable LinksAvailable
       variable LinkInUse

       for {set link 0} {$link < $NLinks} {incr link} {
           if {$link < $LinksAvailable} then {
              if {$LinkInUse($link) == 1} then {
                 MergeCommand 22 "link $link GO" $link 2
              }
           }
        }
}

    load_MergeLinks

