
#include "common.h"
#include "message.h"

#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>


#include "stats.h"


extern STATUS_TABLE *status_table;
extern BUFFER_TABLE *buffer_table;
extern LINK_TABLE *link_table[NUMlinks];

extern int * StatsMem[NUMCOUNTERS];
extern int * RatesMem[NUMCOUNTERS];

extern int * MERGE_Verbose;

#define NUMBUFFERS N_DATA_BUFFERS   /*  from common.h */

#ifdef LINK
int obtain_data_buffer(int link_index)
{
    int i, j;
    unsigned int k=0;

        (*StatsMem[GETBUFFER1])++;
        (*StatsMem[GETBUFFER1 + ((link_index+1)*MAXCOUNTERS)])++;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering obtain_data_buffer for link %d", link_index);
    report_message(MSG_TRACE);

//    scan buffer list for free entry

        j = link_index*N_DATA_BUFFERS;
        for (i = 0; i < NUMBUFFERS; i++) {
            if (buffer_table->buffer_status[j+i] == FREE) {
                buffer_table->buffer_status[j+i] = INUSE;
                buffer_table->buffers_free[link_index]--;

                if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Obtained data buffer (0) for link %d ID %d",link_index, j+i);
                report_message(MSG_TRACE);

                return i;
            }
        }

//    no free buffer at present - wait for one to become available

        (*StatsMem[NOBUFFER1])++;
        (*StatsMem[NOBUFFER1 + ((link_index+1)*MAXCOUNTERS)])++;

        if (*MERGE_Verbose != 0)  sprintf (message_buffer,"No free data buffers for link %d in obtain_data_buffer", link_index);
        report_message(MSG_TRACE);

    for (;;) {

        (*StatsMem[NOBUFFER2])++;
        (*StatsMem[NOBUFFER2 + ((link_index+1)*MAXCOUNTERS)])++;

        usleep(20000);      /*   1/50 sec  */

        j = link_index*N_DATA_BUFFERS;
        for (i = 0; i < NUMBUFFERS; i++) {
            if (buffer_table->buffer_status[j+i] == FREE) {
                buffer_table->buffer_status[j+i] = INUSE;
                buffer_table->buffers_free[link_index]--;

                if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Obtained data buffer (1) for link %d ID %d",link_index, j+i);
                report_message(MSG_TRACE);

                return i;
            }
        }

    }

}
#endif

#ifdef LINK
int queue_data_buffer(int link_index, int i)
{

    int j;
    int rc;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering queue_data_buffer for link %d index %d", link_index, i);
    report_message(MSG_TRACE);

         (*StatsMem[QBUFF])++;
         (*StatsMem[QBUFF+((link_index+1)*MAXCOUNTERS)])++;

    j = link_index*N_DATA_BUFFERS;
    if (buffer_table->buffer_status[j+i] == INUSE) {
        buffer_table->buffer_QID[j+i] = buffer_table->buffer_Q[link_index];  buffer_table->buffer_Q[link_index]++;
        buffer_table->buffer_status[j+i] = QUEUED;

        if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Queued data buffer for link %d ID %d QID=%d Q=%d",link_index, j+i, buffer_table->buffer_QID[j+i], buffer_table->buffer_Q[link_index]);
        report_message(MSG_TRACE);

        rc = 0;
    } else {
        if (*MERGE_Verbose != 0)  sprintf (message_buffer, "Attempt to queue invalid buffer (%d) for link %d %d in queue_data_buffer", i, link_index, buffer_table->buffer_status[j+i]);
        report_message(MSG_TRACE);

         (*StatsMem[QFAIL])++;
         (*StatsMem[QFAIL+((link_index+1)*MAXCOUNTERS)])++;

        rc = -1;
    }
    return rc;

}
#endif

#ifdef MASTER
int release_all_data_buffers()
{
    int i,j;
    int ID;

    for (ID = 0; ID < status_table->num_links; ID++) {

        if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering release_all_data_buffers for link %d", ID);
        report_message(MSG_TRACE);

        if (link_table[ID]->link_state != LINK_ISIDLE) {

           j = ID*N_DATA_BUFFERS;
           buffer_table->buffers_free[ID] = N_DATA_BUFFERS;              /* set all to free */
           for (i = 0; i < N_DATA_BUFFERS; i++) {
              buffer_table->buffer_status[j+i] = FREE;  /* set all to free */
           }
        }
    }

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Exiting release_all_data_buffers");
    report_message(MSG_TRACE);
    return 0;
}
#endif

#ifdef LINK
int release_all_data_buffers(int link_index)
{
    int i,j;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering release_all_data_buffers for link %d", link_index);
    report_message(MSG_TRACE);

    j = link_index*N_DATA_BUFFERS;
    buffer_table->buffers_free[link_index] = N_DATA_BUFFERS;              /* set all to free */
    for (i = 0; i < N_DATA_BUFFERS; i++) {
       buffer_table->buffer_status[j+i] = FREE;  /* set all to free */
    }

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Exiting release_all_data_buffers");
    report_message(MSG_TRACE);
    return 0;
}
#endif

#ifdef MASTER
int report_data_buffer_state()
{
    int i, j;
    int ID;
    unsigned int nfree=0;
    unsigned int ninuse=0;
    unsigned int nqueued=0;
    unsigned int nprocessing=0;
    unsigned int ndone=0;

    for (ID = 0; ID < status_table->num_links; ID++) {

      sprintf (message_buffer,"Entering report_data_buffer_status for link %d", ID);
      report_message(MSG_INFORMATION);

      nfree=ninuse=nqueued=nprocessing=ndone=0;

      if (link_table[ID]->link_state != LINK_ISIDLE) {

        j = ID*N_DATA_BUFFERS;
        for (i = 0; i < N_DATA_BUFFERS; i++) {
            if (buffer_table->buffer_status[j+i] == FREE) {nfree++;}
            if (buffer_table->buffer_status[j+i] == INUSE) {ninuse++;}
            if (buffer_table->buffer_status[j+i] == QUEUED) {nqueued++;}
            if (buffer_table->buffer_status[j+i] == PROCESSING) {nprocessing++;}
            if (buffer_table->buffer_status[j+i] == DONE) {ndone++;}
        }
        
        sprintf (message_buffer,"total number of data buffers %d", N_DATA_BUFFERS);
        report_message(MSG_INFORMATION);
        sprintf (message_buffer,"total number of free data buffers %d", nfree);
        report_message(MSG_INFORMATION);
        sprintf (message_buffer,"total number of inuse data buffers %d", ninuse);
        report_message(MSG_INFORMATION);
        sprintf (message_buffer,"total number of queued data buffers %d", nqueued);
        report_message(MSG_INFORMATION);
        sprintf (message_buffer,"total number of processing data buffers %d", nprocessing);
        report_message(MSG_INFORMATION);
        sprintf (message_buffer,"total number of done data buffers %d", ndone);
        report_message(MSG_INFORMATION);
        sprintf (message_buffer,"Exiting report_data_buffer_status for link %d", ID);
        report_message(MSG_INFORMATION);

      }

    }
}
#endif

#ifdef MERGE
int obtain_data_buffer(int link_index)
{
    int i, j;
    int k, l;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering obtain_data_buffer for link %d", link_index);
//    report_message(MSG_TRACE)

         (*StatsMem[GETBUFFER2])++;
         (*StatsMem[GETBUFFER2+((link_index+1)*MAXCOUNTERS)])++;

//    scan link buffer Q for oldest entry

        k = 0x7fffffff;

        j = link_index*N_DATA_BUFFERS;
        for (i = 0; i < NUMBUFFERS; i++) {
            if (buffer_table->buffer_status[j+i] == QUEUED) {
                if (buffer_table->buffer_QID[j+i] < k) {k = buffer_table->buffer_QID[j+i]; l=i;}
            }
        }
        i = l;    /*  oldest buffer  */

        if (k != 0x7fffffff) {

//    found a buffer 

           buffer_table->buffer_status[j+i] = PROCESSING;

        if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Obtained data buffer for merge link %d ID %d QID=%d",link_index, j+i, buffer_table->buffer_QID[j+i]);
           report_message(MSG_TRACE);

           return i;
        }

//    no buffer at present

         (*StatsMem[NOBUFFER3])++;
         (*StatsMem[NOBUFFER3+((link_index+1)*MAXCOUNTERS)])++;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"No data buffer for link %d pending", link_index);
    report_message(MSG_TRACE);

      return -1;
}
#endif

#ifdef LINK
int release_data_buffer(int link_index, int i)
{
    int rc;
    int j;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering release_data_buffer for link %d with buffer index %d", link_index, i);
    report_message(MSG_TRACE);

         (*StatsMem[RBUFF1])++;
         (*StatsMem[RBUFF1+((link_index+1)*MAXCOUNTERS)])++;

    j = link_index*N_DATA_BUFFERS;
    if (buffer_table->buffer_status[j+i] == DONE) {
        buffer_table->buffer_status[j+i] = FREE;
        buffer_table->buffers_free[link_index]++;

        if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Released data buffer for link %d ID %d QID=%d",link_index, j+i, buffer_table->buffer_QID[j+i]);
        report_message(MSG_TRACE);

        rc = 0;
    } else {
        if (*MERGE_Verbose != 0)  sprintf (message_buffer, "Attempt to release free buffer (%d) in release_data_buffer", i);
        report_message(MSG_TRACE);

         (*StatsMem[RFAIL1])++;
         (*StatsMem[RFAIL1+((link_index+1)*MAXCOUNTERS)])++;

        rc = -1;
    }

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Exiting release_data_buffer (index %d)", i);
    report_message(MSG_TRACE);
    return rc;
}
#endif

#ifdef MERGE
int release_data_buffer(int link_index, int i)
{
    int rc;
    int j;

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Entering release_data_buffer for merge link %d with buffer index %d", link_index, i);
    report_message(MSG_TRACE);

         (*StatsMem[RBUFF2])++;
         (*StatsMem[RBUFF2+((link_index+1)*MAXCOUNTERS)])++;

    j = link_index*N_DATA_BUFFERS;
    if (buffer_table->buffer_status[j+i] == DONE) {
        buffer_table->buffer_status[j+i] = FREE;
        buffer_table->buffers_free[link_index]++;

        if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Released data buffer for link %d ID %d QID=%d",link_index, j+i, buffer_table->buffer_QID[j+i]);
        report_message(MSG_TRACE);

        rc = 0;
    } else {
        if (*MERGE_Verbose != 0)  sprintf (message_buffer, "Attempt to release free buffer (%d) in release_data_buffer", i);
        report_message(MSG_TRACE);

         (*StatsMem[RFAIL2])++;
         (*StatsMem[RFAIL2+((link_index+1)*MAXCOUNTERS)])++;

        rc = -1;
    }

    if (*MERGE_Verbose != 0)  sprintf (message_buffer,"Exiting release_data_buffer (index %d)", i);
    report_message(MSG_TRACE);
    return rc;
}
#endif

