

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "message.h"
#include "buffer.h"
#include "transfer.h"

#include "merge.h"
#include "stats.h"

extern int NumRunningLinks;

extern STATUS_TABLE *status_table;
extern BUFFER_TABLE *buffer_table;
extern LINK_TABLE *link_table[NUMlinks];

extern int * MERGE_RunOptions;
extern int * MERGE_HardwareVersion;

extern int *MERGE_Verbose;

extern int * StatsMem[NUMCOUNTERS];

extern int do_storage (char *, int);

extern  int BufferSize;

extern UINT64 lastTime;      /*  time of last data item  */
extern UINT64 oldestTime;    /*  from last data item  */
extern int oldestLink;       /*  from last data item */
//extern unsigned long wr4863;       /*  from last Info#5 */

extern ItDesc_Ptr OutputBufferPtr;
extern int OutputBufferLen;
extern char * OutputBuffer;

extern int dataStart[NUMlinks];
extern int bufferNumber[NUMlinks];
extern char * dataBuffer[NUMlinks];
extern int dataLen[NUMlinks];



   int do_merge () {

/*
    processes link buffer qs & generates merged data until 1 link q becomes empty
    then returns with that q id
*/

       int done = FALSE;
       int ID, i, j;
       NewItDesc_Ptr ps, os;
       int module, icode, gcode;

       unsigned long wr4863;

       if (*MERGE_Verbose != 0) sprintf(message_buffer, "do_merge merging %d links  0x%016llx  0x%016llx", NumRunningLinks, lastTime, oldestTime);
       report_message(MSG_TRACE);

       while (done == FALSE) {
          os = NULL;

          oldestTime = 0;

         for (ID = 0; ID < status_table->num_links; ID++) {
             if (link_table[ID]->link_state != LINK_GOING) continue;

             ps = (NewItDesc_Ptr)dataBuffer[ID];
             ps += dataStart[ID];

             if (*MERGE_Verbose != 0) sprintf(message_buffer, "merge#1 %d 0x%x 0x%016llx 0x%016llx %d %d", ID, ps, ps->Time, oldestTime, oldestLink, dataStart[ID]);
             report_message(MSG_DEBUG);

             if (ps->Code == WR_CODE) continue;

             if (ps->Time < oldestTime || oldestTime == 0) {
 
                os = ps;                /* this is the next item selected */
                oldestTime = os->Time;  /* and its time */
                oldestLink = ID;

                if (*MERGE_Verbose != 0) sprintf(message_buffer, "merge#2 %d 0x%x 0x%016llx 0x%016llx %d", ID, ps, ps->Time, oldestTime, oldestLink);
                report_message(MSG_DEBUG);

             }
          }     ;//  (ID = 0; ID < status_table->num_links; ID++)  - oldest TS now found

          if (*MERGE_Verbose != 0) sprintf(message_buffer, "merge#3 oldest %d 0x%8x 0x%016llx %d %d",oldestLink,os,oldestTime,dataStart[oldestLink],dataLen[oldestLink]);
          report_message(MSG_DEBUG);

          if (os == NULL) goto L01;  // be cautious - be sure an oldest has really been found

            dataStart[oldestLink]++;

            if (dataStart[oldestLink] >= dataLen[oldestLink]) {
               dataStart[oldestLink] = 0;

               if (*MERGE_Verbose != 0) sprintf(message_buffer, "merge_server (merge done=TRUE) %d", oldestLink);
               report_message(MSG_DEBUG);
 
               i = bufferNumber[oldestLink];
               j = oldestLink*N_DATA_BUFFERS;

               buffer_table->buffer_status[i+j] = DONE;            //  flag all data in this buffer as been processed
               done = TRUE;     /*  buffer complete  */
            }

    switch (*MERGE_HardwareVersion) {
        case  0: 

#include "mergeAIDA.c"
	  
	   break;
	  
	case  1:
	  
#include "mergeR3B.c"

            break;
	    
        default:

            printf ("Unsupported VERSION %s in Merger Link handle_buffer\n",*MERGE_HardwareVersion);
            exit(32);

            break;
     }     /*   end of switch  */

L01:       continue;
      } // end of     while (done == FALSE)    loop - one of the input buffers is empty

       if (*MERGE_Verbose != 0) sprintf(message_buffer, "do_merge end of buffer link %d", oldestLink);
       report_message(MSG_TRACE);

      return oldestLink;
   }




